/*
 * (C) Copyright 2000 - 2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Hacked for MPC8260 by Murray.Jensen@cmst.csiro.au, 19-Oct-00, with
 * changes based on the file arch/ppc/mbxboot/m8260_tty.c from the
 * Linux/PPC sources (m8260_tty.c had no copyright info in it).
 *
 * Martin Krause, 8 Jun 2006
 * Added CONFIG_SERIAL_MULTI support
 */

/*
 * Minimal serial functions needed to use one of the PSC ports
 * as serial console interface.
 */

#include <common.h>
#include <asm/leon.h>
#include <asm/leon3.h>
#include <ambapp.h>

//#define CACHE_BYPASS 1

DECLARE_GLOBAL_DATA_PTR;

extern LEON3_APBUART_Regs_Map *leon3_apbuart;

int serial_init (void)
{
	ambapp_apbdev apbdev;
	
	/* find UART */
	if ( ambapp_apb_first(VENDOR_GAISLER,GAISLER_APBUART,&apbdev) == 1 ){
		
		leon3_apbuart = (LEON3_APBUART_Regs_Map *)apbdev.address;
		
		/* found apbuart, let's init...
		 *
		 * Set scaler / baud rate
		 *
		 * Receiver & transmitter enable
		 */
		leon3_apbuart->scaler = CFG_GRLIB_APBUART_SCALER;
		
		leon3_apbuart->ctrl = (LEON_REG_UART_CTRL_RE | LEON_REG_UART_CTRL_TE);
		
		return 0;
	}
	return -1; /* didn't find hardware */
}

void serial_putc(const char c)
{
	if (c == '\n')
		serial_putc_raw('\r');

	serial_putc_raw(c);
}

void serial_putc_raw(const char c)
{
	if ( !leon3_apbuart )
		return;
	
	/* Wait for last character to go. */
#ifdef CACHE_BYPASS	
	while (!( leon_load_reg_cachemiss(&leon3_apbuart->status) & LEON_REG_UART_STATUS_THE))
#else
	while (!(leon3_apbuart->status & LEON_REG_UART_STATUS_THE))
#endif	
		;
	
	/* Send data */
	leon3_apbuart->data = c;
	
#ifdef LEON_DEBUG
	/* Wait for data to be sent */
#ifdef CACHE_BYPASS
	while (!(leon_load_reg_cachemiss(&leon3_apbuart->status) & LEON_REG_UART_STATUS_TSE))
#else
	while (!(leon3_apbuart->status & LEON_REG_UART_STATUS_TSE))
#endif
		;
	
#endif
}


void serial_puts (const char *s)
{
	while (*s) {
		serial_putc (*s++);
	}
}

int serial_getc(void)
{
	if ( !leon3_apbuart )
		return 0;
	
	/* Wait for a character to arrive. */
#ifdef CACHE_BYPASS
	while (!(leon_load_reg_cachemiss(&leon3_apbuart->status) & LEON_REG_UART_STATUS_DR))
#else
	while (!(leon3_apbuart->status & LEON_REG_UART_STATUS_DR))
#endif
		;

	/* read data */
#ifdef CACHE_BYPASS	
	return leon_load_reg_cachemiss(&leon3_apbuart->data);
#else
	return leon3_apbuart->data;
#endif
}

int serial_tstc(void)
{
	if ( leon3_apbuart )
#ifdef CACHE_BYPASS	
		return (leon_load_reg_cachemiss(&leon3_apbuart->status) & LEON_REG_UART_STATUS_DR);
#else
		return (leon3_apbuart->status & LEON_REG_UART_STATUS_DR);
#endif
	return 0;
}

/* set baud rate generator */
void serial_setbrg (void)
{
	/* update baud rate settings, read it from gd->baudrate */
	unsigned int scaler;
	if ( leon3_apbuart && (gd->baudrate>0) ){
		scaler=(((CONFIG_SYS_CLK_FREQ*10)/(gd->baudrate*8))-5)/10;
		leon3_apbuart->scaler = scaler;
	}
	return;
}
