#ifndef __LEON3_H__
#define __LEON3_H__

/* LEON 3 Header file
 *
 */

#ifdef CONFIG_LEON3

#define ASI_LEON3_CACHEMISS 1
#define ASI_LEON3_SYSCTRL   0x02
#define ASI_LEON3_IFLUSH		0x10
#define ASI_LEON3_DFLUSH    0x11

#define ASI_LEON_PCI		0x04
#define ASI_LEON_IFLUSH		0x05
#define ASI_LEON_DFLUSH		0x06
#define ASI_LEON_ITAG		0x0c
#define ASI_LEON_IDATA		0x0d
#define ASI_LEON_DTAG		0x0e
#define ASI_LEON_DDATA		0x0f
#define ASI_LEON_MMUFLUSH	0x18
#define ASI_LEON_MMUREGS	0x19
#define ASI_LEON_BYPASS		0x1c
#define ASI_LEON_FLUSH_PAGE	0x10

/*
 *  The following defines the bits in the LEON UART Status Registers.
 */

#define LEON_REG_UART_STATUS_DR   0x00000001	/* Data Ready */
#define LEON_REG_UART_STATUS_TSE  0x00000002	/* TX Send Register Empty */
#define LEON_REG_UART_STATUS_THE  0x00000004	/* TX Hold Register Empty */
#define LEON_REG_UART_STATUS_BR   0x00000008	/* Break Error */
#define LEON_REG_UART_STATUS_OE   0x00000010	/* RX Overrun Error */
#define LEON_REG_UART_STATUS_PE   0x00000020	/* RX Parity Error */
#define LEON_REG_UART_STATUS_FE   0x00000040	/* RX Framing Error */
#define LEON_REG_UART_STATUS_ERR  0x00000078	/* Error Mask */

/*
 *  The following defines the bits in the LEON UART Ctrl Registers.
 */

#define LEON_REG_UART_CTRL_RE     0x00000001	/* Receiver enable */
#define LEON_REG_UART_CTRL_TE     0x00000002	/* Transmitter enable */
#define LEON_REG_UART_CTRL_RI     0x00000004	/* Receiver interrupt enable */
#define LEON_REG_UART_CTRL_TI     0x00000008	/* Transmitter interrupt enable */
#define LEON_REG_UART_CTRL_PS     0x00000010	/* Parity select */
#define LEON_REG_UART_CTRL_PE     0x00000020	/* Parity enable */
#define LEON_REG_UART_CTRL_FL     0x00000040	/* Flow control enable */
#define LEON_REG_UART_CTRL_LB     0x00000080	/* Loop Back enable */

#define LEON3_GPTIMER_EN 1
#define LEON3_GPTIMER_RL 2
#define LEON3_GPTIMER_LD 4
#define LEON3_GPTIMER_IRQEN 8

/*
 *  The following defines the bits in the LEON PS/2 Status Registers.
 */

#define LEON_REG_PS2_STATUS_DR   0x00000001 /* Data Ready */
#define LEON_REG_PS2_STATUS_PE   0x00000002 /* Parity error */
#define LEON_REG_PS2_STATUS_FE   0x00000004 /* Framing error */
#define LEON_REG_PS2_STATUS_KI   0x00000008 /* Keyboard inhibit */
 

/*
 *  The following defines the bits in the LEON PS/2 Ctrl Registers.
 */

#define LEON_REG_PS2_CTRL_RE     0x00000001 /* Receiver enable */
#define LEON_REG_PS2_CTRL_TE     0x00000002 /* Transmitter enable */
#define LEON_REG_PS2_CTRL_RI     0x00000004 /* Keyboard receive interrupt  */
#define LEON_REG_PS2_CTRL_TI     0x00000008 /* Keyboard transmit interrupt */

#ifndef __ASSEMBLY__
/* do a cache miss load */

static __inline__ unsigned long long leon_load_reg_cachemiss_qword(unsigned long paddr)
{
	unsigned long long retval;
	__asm__ __volatile__("ldda [%1] %2, %0\n\t":
			     "=r"(retval): "r"(paddr), "i"(ASI_LEON3_CACHEMISS));
	return retval;
}

static __inline__ unsigned long leon_load_reg_cachemiss(unsigned long paddr)
{
	unsigned long retval;
	__asm__ __volatile__("lda [%1] %2, %0\n\t":
			     "=r"(retval): "r"(paddr), "i"(ASI_LEON3_CACHEMISS));
	return retval;
}

static __inline__ unsigned short leon_load_reg_cachemiss_word(unsigned long paddr)
{
	unsigned short retval;
	__asm__ __volatile__("lduha [%1] %2, %0\n\t":
			     "=r"(retval): "r"(paddr), "i"(ASI_LEON3_CACHEMISS));
	return retval;
}

static __inline__ unsigned char leon_load_reg_cachemiss_byte(unsigned long paddr)
{
	unsigned char retval;
	__asm__ __volatile__("lduba [%1] %2, %0\n\t":
			     "=r"(retval): "r"(paddr), "i"(ASI_LEON3_CACHEMISS));
	return retval;
}

/* do a physical address bypass write, i.e. for 0x80000000 */
static __inline__ void leon_store_reg_prom(unsigned long paddr, unsigned long value)
{
	__asm__ __volatile__("sta %0, [%1] %2\n\t"::"r"(value), "r"(paddr),
			     "i"(ASI_LEON_BYPASS):"memory");
}

/* do a physical address bypass load, i.e. for 0x80000000 */
static __inline__ unsigned long leon_load_reg_prom(unsigned long paddr)
{
	unsigned long retval;
	__asm__ __volatile__("lda [%1] %2, %0\n\t":
			     "=r"(retval): "r"(paddr), "i"(ASI_LEON_BYPASS));
	return retval;
}


static __inline__ void leon_or_reg_cachemiss(unsigned long paddr,unsigned long value) {
  unsigned long v = leon_load_reg_cachemiss(paddr);
  *((unsigned long *)paddr) = v | value;
}

static __inline__ void leon_and_reg_cachemiss(unsigned long paddr,unsigned long value) {
  unsigned long v = leon_load_reg_cachemiss(paddr);
  *((unsigned long *)paddr) = v & value;
}

/* flush cache */
static __inline__ void leon3_dcache_flush(void) {
  __asm__ __volatile__("sta %%g0, [%%g0] %0\n\t": :
                       "i" (ASI_LEON3_DFLUSH) : "memory");
}

typedef struct {
	volatile unsigned int ilevel;
	volatile unsigned int ipend;
	volatile unsigned int iforce;
	volatile unsigned int iclear;
	volatile unsigned int notused00;
	volatile unsigned int notused01;
	volatile unsigned int notused02;
	volatile unsigned int notused03;
	volatile unsigned int notused10;
	volatile unsigned int notused11;
	volatile unsigned int notused12;
	volatile unsigned int notused13;
	volatile unsigned int notused20;
	volatile unsigned int notused21;
	volatile unsigned int notused22;
	volatile unsigned int notused23;
	volatile unsigned int mask[16];
} LEON3_IrqCtrl_Regs_Map;

typedef struct {
	volatile unsigned int data;
	volatile unsigned int status;
	volatile unsigned int ctrl;
	volatile unsigned int scaler;
} LEON3_APBUART_Regs_Map;

typedef struct {
	volatile unsigned int val;
	volatile unsigned int rld;
	volatile unsigned int ctrl;
	volatile unsigned int unused;
} LEON3_GpTimerElem_Regs_Map;

#define LEON3_GPTIMER_CTRL_EN	0x1	/* Timer enable */
#define LEON3_GPTIMER_CTRL_RS	0x2	/* Timer reStart  */
#define LEON3_GPTIMER_CTRL_LD	0x4	/* Timer reLoad */
#define LEON3_GPTIMER_CTRL_IE	0x8	/* interrupt enable */
#define LEON3_GPTIMER_CTRL_IP	0x10	/* interrupt flag/pending */
#define LEON3_GPTIMER_CTRL_CH	0x20	/* Chain with previous timer */


typedef struct {
	volatile unsigned int scalar;
	volatile unsigned int scalar_reload;
	volatile unsigned int config;
	volatile unsigned int unused;
	volatile LEON3_GpTimerElem_Regs_Map e[8];
} LEON3_GpTimer_Regs_Map;

typedef struct {
  volatile unsigned int iodata;
  volatile unsigned int ioout;
  volatile unsigned int iodir;
  volatile unsigned int irqmask;
  volatile unsigned int irqpol;
  volatile unsigned int irqedge;
} LEON3_IOPORT_Regs_Map;

typedef struct {
  volatile unsigned int write;
  volatile unsigned int dummy;
  volatile unsigned int txcolor;
  volatile unsigned int bgcolor;
} LEON3_VGA_Regs_Map;
extern volatile LEON3_VGA_Regs_Map *leon_apbvga;

typedef struct {
  volatile unsigned int data;
  volatile unsigned int status;
  volatile unsigned int ctrl;
} LEON3_APBPS2_REGS_Map;
extern volatile LEON3_APBPS2_REGS_Map *leon_apbps2;

typedef struct {
	unsigned int mcfg1,mcfg2,mcfg3;
} LEON3_MCTRL;

typedef struct {
	unsigned int sdcfg;
} LEON3_SDCTRL;

enum {
	LEON_MEMCTRL_NONE = 0,
	LEON_MEMCTRL_MCTRL = 1,
	LEON_MEMCTRL_SDCTRL = 2,
	LEON_MEMCTRL_SRAM = 3
};

typedef struct {
	int ctrltype;
	union {
		LEON3_MCTRL *mctrl;
		LEON3_SDCTRL *sdctrl;
	} mbase;
} LEON3_MEMCTRL;



extern void vendor_dev_string(unsigned long conf, char *vendorbuf, char *devbuf);

#endif /* __ASSEMBLY__ */

#endif
#endif

